EventStore v0.12.0 EventStore.RecordedEvent View Source
EventStore.RecordedEvent contains the persisted data and metadata for a single event.
Events are immutable once recorded.
Recorded event fields
event_id- a globally unique, monotonically incrementing and gapless integerstream_uuid- the stream identity for the eventstream_version- the version of the stream for the eventcorrelation_id- an optional identifier used to correlate related messagescausation_id- an optional identifier used to identify which message you are responding todata- the serialized event as binary datametadata- the serialized event metadata as binary datacreated_at- the date/time, in UTC, indicating when the event was created
Link to this section Summary
Link to this section Types
Link to this type
t()
t() :: %EventStore.RecordedEvent{causation_id: String.t(), correlation_id: String.t(), created_at: NaiveDateTime.t(), data: binary(), event_id: non_neg_integer(), event_type: String.t(), metadata: binary(), stream_uuid: String.t(), stream_version: non_neg_integer()}